<?php class shibboleth {

    const USER_LINKED = "user_linked";
    const USER_CREATED = "user_created";
    const USER_CONFLICT = "user_conflict";

    public static function getIdentityProviderEntityId(){
        if(isset($_SERVER['Shib-Identity-Provider'])){
            return $_SERVER['Shib-Identity-Provider'];
        }
        return null;
    }

    public static function getIdentityProviders(){
        $path=rtrim($_SERVER['DOCUMENT_ROOT'],'/').'/conf/identityProviders.json';
        if(file_exists($path) && $content=@file_get_contents($path)){
            $content=json_decode($content,true);
            if($content && isset($content['identityProviders'])){ return $content['identityProviders']; }
        }
        return [];
    }

    public static function isAuthenticated(){
        //Is there a SAML entityID in the server variables?
        $entityId=static::getIdentityProviderEntityId();
        if(empty($entityId)){ return false; }
        //Is it one of the entityIDs recognised by this IceBear install?
        $providers=static::getIdentityProviders();
        if(empty($providers)){ return false; }
        foreach($providers as $idp){
            if($entityId===$idp['entityID']){
                return true;
            }
        }
        return false;
    }

    /**
     * @throws ServerException
     */
    private static function getSamlUserAttributes(){
        if(!static::isAuthenticated()){ return null; }
        $attributes=['mail','eppn','givenName','fullName','familyName','displayName','orcid'];
        $user=[];
        foreach($attributes as $attr){
            $value=static::getSingleValueFromAttribute($attr);
            if(!empty($value)){ $user[$attr]=$value; }
        }
        $aliases=[
            'email'=>'mail',
            'eduPersonPrincipalName'=>'eppn',
            'firstName'=>'givenName',
            'lastName'=>'familyName',
            'surname'=>'familyName'
        ];
        foreach($aliases as $alias=>$attribute){
            if(!isset($user[$attribute])){
                $value=static::getSingleValueFromAttribute($alias);
                if(!empty($value)){ $user[$attribute]=$value; }
            }
        }
        return $user;
    }

    /**
     * @throws ServerException
     */
    private static function getSingleValueFromAttribute($attribute){
        if(!isset($_SERVER[$attribute])){ return null; }
        $parts=explode(';', $_SERVER[$attribute]);
        if(1===count($parts)){ return $parts[0]; }
        for($i=1;$i<count($parts);$i++){
            if($parts[0]!==$parts[$i]){
                throw new ServerException("Multiple and different values for single-value SAML attribute $attribute: [".implode(';',$parts)."]");
            }
        }
        return $parts[0];
    }

    /**
     * @throws \Random\RandomException
     * @throws NotFoundException
     * @throws ForbiddenException
     * @throws BadRequestException
     * @throws ServerException
     */
    public static function getIceBearUser(){
        $samlUser=static::getSamlUserAttributes();
        $nameMatch=null;
        $emailMatch=null;
        $ret=null;
        session::becomeAdmin();
        if(isset($samlUser['eppn'])){
            $nameMatch=user::getByProperty('edupersonprincipalname',$samlUser['eppn']);
            if($nameMatch){ $nameMatch=$nameMatch['rows'][0]; }
        }
        if(isset($samlUser['mail'])){
            $emailMatch=user::getByProperty('email',$samlUser['mail']);
            if($emailMatch){ $emailMatch=$emailMatch['rows'][0]; }
        } else {
            throw new ServerException('Expected SAML user to have attribute [mail] but none found');
        }
        if($nameMatch && $emailMatch && $nameMatch['id']!==$emailMatch['id']){

            //SAML email and eduPersonPrincipalName refer to different IceBear users!
            $ret=static::USER_CONFLICT;

        } else if($nameMatch && $emailMatch){

            //Matched on both EPPN and email, great
            $ret=$nameMatch;

        } else if($nameMatch){

            //Matched on SAML eduPersonPrincipalName but email is different. Don't mess with the
            //IceBear user's email, it might affect, for example, Formulatrix import, causing duplicate users.
            $ret=$nameMatch;

        } else if($emailMatch){

            //Matched on email address, but account doesn't have this eduPersonPrincipalName (and nor does any
            //other account, otherwise we would have had conflict). First check that the IceBear account does
            //NOT already have an eduPersonPrincipalName...
            if (!empty($emailMatch['edupersonprincipalname'])){
                $ret = static::USER_CONFLICT;
            } else if (isset($samlUser['eppn'])){
                //Would have matched, if a user had this eduPersonPrincipalName
                //So we set it to this account.
                user::update($emailMatch['id'], ['edupersonprincipalname' => $samlUser['eppn']]);
                $ret = static::USER_LINKED;
            } else {
                $ret = $emailMatch;
            }

        } else {

            //No match on EPPN or email. Create an IceBear user.
            $iceBearUser=[
                'email'=>$samlUser['mail'],
                'password'=>random_bytes(32) //never using this, but database complains without it,
            ];

            //Set the username
            if(isset($samlUser['eppn'])){
                $iceBearUser['name']=$samlUser['eppn'];
            } else {
                $iceBearUser['name']=$samlUser['mail'];
            }

            //Set the EPPN if we have it
            if(isset($samlUser['eppn'])){
                $iceBearUser['edupersonprincipalname']=$samlUser['eppn'];
            }

            //Set the user's "full name"
            if(isset($samlUser['displayName'])){
                $iceBearUser['fullname']=$samlUser['displayName'];
            } else if(isset($samlUser['commonName'])){
                $iceBearUser['fullname']=$samlUser['commonName'];
            } else if(isset($samlUser['fullName'])){
                $iceBearUser['fullname']=$samlUser['fullName'];
            } else if(isset($samlUser['givenName']) && isset($samlUser['familyName'])){
                $iceBearUser['fullname']=$samlUser['familyName'].', '.$samlUser['givenName'];
            } else {
                throw new ServerException('Cannot create IceBear user - no usable name in SAML attributes');
            }

            user::create($iceBearUser);
            $ret=static::USER_CREATED;

        }
        session::revertAdmin();
        return $ret;
    }

}